// -*- tab-width: 4; indent-tabs-mode: nil; c-basic-offset: 2 -*-
// vi: set et ts=4 sw=2 sts=2:
#ifndef DUNE_CURVED_SURFACE_GRID_CONCEPTS_HH
#define DUNE_CURVED_SURFACE_GRID_CONCEPTS_HH

#include <dune/common/concept.hh>

namespace Dune {
namespace Concept {

//! Concept: objects that can be called with given argument list `Args...`
template< class... Args >
struct Callable
{
  template< class F >
  auto require (F&& f) -> decltype(
    f(std::declval<Args>()...)
  );
};

//! Check if `F` models the `Callable` concept with the given arguments `Args...`
template< class F, class... Args >
constexpr auto isCallable ()
{ return models<Concept::Callable<Args...>, F>(); }


//! Concept: function associated to the `LocalContext` that can be evaluated in local coordinates
template< class LocalContext >
struct LocalFunction
{
  using LocalCoordinate = typename LocalContext::Geometry::LocalCoordinate;

  template< class LF >
  auto require (LF&& lf) -> decltype(
    lf.bind(std::declval<LocalContext>()),
    lf.unbind(),
    lf.localContext(),
    requireConcept<Concept::Callable<LocalCoordinate>>(lf),
    requireConvertible<LocalContext>(lf.localContext())
  );
};

//! Check if `LF` models the `LocalFunction` concept on the given `LocalContext`
template< class LF, class LocalContext >
constexpr bool isLocalFunction ()
{ return models<Concept::LocalFunction<LocalContext>, LF>(); }


//! Concept: function associated to the `Grid` that can be evaluated in global coordinates
template< class Grid >
struct GridFunction
{
  using LocalContext = typename Grid::template Codim<0>::Entity;
  using GlobalCoordinate = typename LocalContext::Geometry::GlobalCoordinate;

  template< class GF >
  auto require (GF&& gf) -> decltype(
    localFunction(gf),
    gf.entitySet(),
    requireConcept<Concept::Callable<GlobalCoordinate>>(gf),
    requireConcept<Concept::LocalFunction<LocalContext>>(localFunction(gf))
  );
};

//! Check if `GF` models the `GridFunction` concept on the given `Grid`
template< class GF, class Grid >
constexpr bool isGridFunction ()
{ return models<Concept::GridFunction<Grid>, GF>(); }

} // end namespace Concept
} // end namespace Dune

#endif // DUNE_CURVED_SURFACE_GRID_CONCEPTS_HH